# Interface implementation Ball And Beam ====================================================

# The ball and beam can be used in two modes, either just the Beam, in which case there is a
# single output (measurement signal) or the BallAndBeam, in which case there are two.
# There are a few union types defined for convenience, these are
# AbstractBeam = Union{Beam, BeamSimulator}
# AbstractBallAndBeam   = Union{BallAndBeam, BallAndBeamSimulator}
# AbstractBeamOrBallAndBeam = All types
# Although not Abstract per se, the names AbstractBeam etc. were chosen since this reflects
# their usage in dispatch.

export Beam, BeamSimulator, AbstractBeam, BallAndBeam, BallAndBeamSimulator, AbstractBeamOrBallAndBeam

struct Beam <: PhysicalProcess
    h::Float64
    bias::Float64
end
Beam() = Beam(0.01, 0.0)

include("define_beam_system.jl")
const beam_system, nice_beam_controller = define_beam_system()
# nice_beam_controller gives ϕₘ=56°, Aₘ=4, Mₛ = 1.6. Don't forget to discretize it before use
struct BeamSimulator <: SimulatedProcess
    h::Float64
    state::Vector{Float64} # states defined by the file define_beam_system
    sys::StateSpace
    BeamSimulator() = new(0.01, init_sysfilter(beam_system), c2d(beam_system, 0.01)[1])
    BeamSimulator(h::Real) = new(Float64(h), init_sysfilter(beam_system), c2d(beam_system, h)[1])
end

struct BallAndBeam <: PhysicalProcess
    h::Float64
    bias::Float64
end
BallAndBeam() = BallAndBeam(0.01, 0.0)

struct BallAndBeamSimulator <: SimulatedProcess
    h::Float64
    state::Vector{Float64} # pos,vel,angle,ω
end
BallAndBeamSimulator() = BallAndBeamSimulator(0.01, zeros(4))


const AbstractBeam              = Union{Beam, BeamSimulator}
const AbstractBallAndBeam       = Union{BallAndBeam, BallAndBeamSimulator}
const AbstractBeamOrBallAndBeam = Union{AbstractBeam, AbstractBallAndBeam}


num_outputs(p::AbstractBeam)             = 1
num_outputs(p::AbstractBallAndBeam)      = 2
num_inputs(p::AbstractBeamOrBallAndBeam) = 1
outputrange(p::AbstractBeam)             = [(-10,10)]
outputrange(p::AbstractBallAndBeam)      = [(-10,10),(-1,1)] # Beam angle, Ball position
inputrange(p::AbstractBeamOrBallAndBeam) = [(-10,10)]
isstable(p::AbstractBeam)                = true
isstable(p::AbstractBallAndBeam)         = false
isasstable(p::AbstractBeamOrBallAndBeam) = false
sampletime(p::AbstractBeamOrBallAndBeam) = p.h
bias(p::AbstractBeamOrBallAndBeam)       = p.bias
bias(p::BeamSimulator)                   = 0
bias(p::BallAndBeamSimulator)            = 0


control(p::AbstractBeamOrBallAndBeam, u) = ccall((:comedi_write, comedipath),Int32,
                                        (Int32,Int32,Int32,Int32),0,1,1,num2io(u[1]+p.bias))

function control(p::BeamSimulator, u)
    sysfilter!(p.state, p.sys, u)
end
control(p::BallAndBeamSimulator, u)  = error("Not yet implemented")

measure(p::Beam) = io2num(ccall((:comedi_read,comedipath), Int32,
                                        (Int32,Int32,Int32), 0,0,0))
measure(p::BallAndBeam) = [io2num(ccall((:comedi_read,comedipath), Int32,
(Int32,Int32,Int32), 0,0,i)) for i = 0:1]

measure(p::BeamSimulator)        = vecdot(p.sys.C,p.state)
measure(p::BallAndBeamSimulator) = error("Not yet implemented")


const comedipath = Pkg.dir("LabProcesses","c","comedi_bridge.so")
const conversion = 65535/20
io2num(x) = x/conversion -10            # Converts from io to float
num2io(x) = round(Int32,x*100 + 2048)   # Converts from regular number to io

initialize(p::AbstractBeamOrBallAndBeam) = ccall((:comedi_start, comedipath),Int32,(Int32,), 0)
finalize(p::AbstractBeamOrBallAndBeam)   = (control(p,0);ccall((:comedi_stop, comedipath),Int32,(Int32,), 0))
initialize(p::BallAndBeamSimulator)      = nothing
finalize(p::BallAndBeamSimulator)        = nothing
initialize(p::BeamSimulator)             = p.state .*= 0
finalize(p::BeamSimulator)               = nothing
